import { by, element, ElementFinder, browser } from "protractor";

export class CalculatorPage {

    getInputLabels() {
        return element.all(by.css("ngparam-input input:not([disabled]) label"));
    }

    getParamInputs() {
        return element.all(by.css("ngparam-input input.form-control"));
    }

    async getParamInputsHavingCalcMode() {
        const ret = [];
        const inputs = this.getParamInputs();
        await inputs.each(async (i) => {
            if (await this.inputHasCalcModeButton(i)) {
                ret.push(i);
            }
        });
        return ret;
    }

    getHeader1() {
        return element(by.css("h1"));
    }

    getSelectById(id: string) {
        return element(by.id(id));
    }

    async getSelectValueText(select: ElementFinder) {
        return await select.element(by.css(".mat-select-value-text > span")).getText();
    }

    async isSelectEmpty(select: ElementFinder) {
        try {
            const text = select.element(by.css(".mat-select-value-text > span"));
            await text.getAttribute("outerHTML"); // await anything trigger the error
            return false;
        } catch (e) { // should be NoSuchElementError
            return true;
        }
    }

    getInputById(id: string) {
        return element(by.id(id));
    }

    getSaveSessionButton() {
        return element(by.css("dialog-save-session button[type=submit]"));
    }

    getCalculateButton() {
        return element(by.css("button#trigger-calculate"));
    }

    getGeneratePabButton() {
        return element(by.css("button#generate-pab"));
    }

    getCheckedCalcModeButtons() {
        return element.all(by.css(`mat-button-toggle.radio_cal[ng-reflect-checked="true"]`));
    }

    getAddStructureButton() {
        return element(by.css("structure-fieldset-container .hyd-window-btns button.add-structure"));
    }

    getCopyStructureButton() {
        return element(by.css("structure-fieldset-container .hyd-window-btns button.copy-structure"));
    }

    getAllLinkButtons() {
        return element.all(by.css("mat-button-toggle.radio_link"));
    }

    getPabResultsTable() {
        return element(by.css(".pab-results-table-inner-container table"));
    }

    getVariatedResultsTable() {
        return element(by.css(".var-results-inner-container table"));
    }

    getAllVariatedResultsTableHeaders() {
        return element.all(by.css("fixedvar-results var-results table thead th"));
    }

    getAllVariatedResultsRows() {
        return element.all(by.css("fixedvar-results var-results table tbody tr"));
    }

    scrollTo(elt: ElementFinder) {
        browser.controlFlow().execute(function () {
            browser.executeScript("arguments[0].scrollIntoView(true)", elt.getWebElement());
        });
    }

    getFixedResultsTable() {
        return element(by.css(".fixed-results-inner-container table"));
    }

    getAllFixedResultsRows() {
        return element.all(by.css("fixed-results table tbody tr"));
    }

    /** return nth <tr> of given <table>, starting at 1 */
    getNthRow(table: ElementFinder, n: number) {
        return table.element(by.css("tbody > tr:nth-of-type(" + n + ")"));
    }

    /** return nth <td> of given <tr>, starting at 1 */
    getNthColumn(tr: ElementFinder, n: number) {
        return tr.element(by.css("td:nth-of-type(" + n + ")"));
    }

    async inputHasCalcModeButton(input: ElementFinder) {
        // get parent (div.container)
        const container = await this.findParentContainer(input);
        // find radio buttons
        const button: ElementFinder = container.element(by.css("mat-button-toggle.radio_cal > button"));
        return await button.isPresent();
    }

    /**
     * @returns true if "fixed mode" button linked to an input is selected
     */
    async inputIsInFixedMode(input: ElementFinder): Promise<boolean> {
        // get parent (div.container)
        const container = await this.findParentContainer(input);
        // find fixed radio button
        const button: ElementFinder = container.element(by.css("mat-button-toggle.radio_fix"));
        return await button.getAttribute("ng-reflect-checked") === "true";
    }

    /**
     * @returns true if "calculated mode" button linked to an input is selected
     */
    async inputIsInCalculatedMode(input: ElementFinder): Promise<boolean> {
        // get parent (div.container)
        const container = await this.findParentContainer(input);
        // find calc radio button
        const button: ElementFinder = container.element(by.css("mat-button-toggle.radio_cal"));
        return await button.getAttribute("ng-reflect-checked") === "true";
    }

    async hasResults() {
        return (
            await this.presentAndVisible("fixedvar-results fixed-results > .fixed-results-container")
            ||
            await this.presentAndVisible("fixedvar-results results-chart > chart-results-container")
            ||
            await this.presentAndVisible("section-results fixed-results > .fixed-results-container")
            ||
            await this.presentAndVisible("remous-results #main-chart")
            ||
            await this.presentAndVisible("pab-results pab-results-table")
            ||
            await this.presentAndVisible("pb-results pb-results-table")
            ||
            await this.presentAndVisible("macrorugo-compound-results macrorugo-compound-results-table")
            ||
            await this.presentAndVisible("jet-results .fixed-results-container")
        );
    }

    async presentAndVisible(selector: string): Promise<boolean> {
        const elt = element(by.css(selector));
        return await elt.isPresent() && await elt.isDisplayed();
    }

    /**
     * For a given <table> element, check that values of all cells of all rows in <tbody> are
     * different from "NaN", "ERR" and optionally ""
     * @param table a <table> ElementFinder
     * @param allowEmpty if true, empty "" values will be considered valid
     */
    async allRowsHaveValidResults(table: ElementFinder, allowEmpty: boolean = false): Promise<boolean> {
        let n = 0;
        let ok = true;
        const invalidValues = ["ERR", "NaN"];
        if (!allowEmpty) {
            invalidValues.push("");
        }

        const nbCols = await table.all(by.css("thead th")).count();
        const rows = await table.all(by.css("tbody tr"));

        for (const row of rows) {
            for (let i = 0; i < nbCols; i++) {
                // get i_th column of n_th row
                const val = await row.element(by.css("td:nth-of-type(" + (i + 1) + ")")).getText();
                // console.log(`TABLE VAL ${n}/${i}=>`, val);
                ok = ok && !invalidValues.includes(val);
            }
            n++;
        }
        return ok;
    }

    /**
     * Returns true if this.hasResults() is true, and if results table are
     * not empty and contain only valid values (no "NaN", no "", no "ERR")
     */
    async hasValidResults(): Promise<boolean> {
        let ok = false;
        if (await this.hasResults()) {
            ok = true;
            // check fixed results
            const frt = this.getFixedResultsTable();
            if (await frt.isPresent() && await frt.isDisplayed()) {
                ok = ok && await this.allRowsHaveValidResults(frt);
            }
            // check variated results
            const vrt = this.getVariatedResultsTable();
            if (await vrt.isPresent() && await vrt.isDisplayed()) {
                ok = ok && await this.allRowsHaveValidResults(vrt);
            }
            // check PAB results
            const prt = this.getPabResultsTable();
            if (await prt.isPresent() && await prt.isDisplayed()) {
                ok = ok && await this.allRowsHaveValidResults(prt, true);
            }
        }
        return ok;
    }

    async hasLog() {
        return await this.nbLogEntries() > 0;
    }

    async nbLogEntries() {
        return await element.all(by.css("log-entry")).count();
    }

    async clickSaveCalcButton() {
        return await element(by.css("#save-calc")).click();
    }

    async clickCloneCalcButton() {
        const cloneButton = element(by.css("#clone-calc"));
        return await cloneButton.click();
    }

    // check that "compute" button is in given enabled/disabled state
    checkCalcButtonEnabled(enabled: boolean) {
        const calcButton = this.getCalculateButton();
        expect(calcButton.isEnabled()).toBe(enabled);
        return calcButton;
    }

    async changeSelectValue(elt: ElementFinder, index: number) {
        await elt.click();
        const optionId = ".cdk-overlay-container mat-option:nth-of-type(" + (index + 1) + ")";
        const option = element(by.css(optionId));
        await option.click();
    }

    // find parent element of elt having class "container"
    async findParentContainer(elt: ElementFinder): Promise<ElementFinder> {
        let i = 8; // garde fous
        while ((await elt.getAttribute("class") !== "container") && (i >= 0)) {
            elt = elt.element(by.xpath(".."));
            i--;
        }
        return elt;
    }

    /**
     * @param elt an <input> element
     * @param mode "fix", "var", "cal" or "link"
     */
    async setParamMode(elt: ElementFinder, mode: string) {
        // get parent (div.container)
        const container = await this.findParentContainer(elt);
        // find radio buttons
        const button = container.element(by.css("mat-button-toggle.radio_" + mode + " > button"));
        await browser.executeScript("window.scrollTo(0, 0);"); // sometimes button slides behind navbar and click() fails
        await button.click();
        // for "var" mode, close the modal
        if (mode === "var") {
            await browser.sleep(500); // wait for the modal to appear
            await element(by.css("dialog-edit-param-values .mat-dialog-actions button")).click();
            await browser.sleep(500); // wait for the navbar to reappear after modal dismissal
        }
    }

    /**
     * @param elt an <input> element
     */
    async getLinkedValueSelect(elt: ElementFinder): Promise<ElementFinder> {
        // get parent (div.container)
        const container = await this.findParentContainer(elt) as ElementFinder;
        // find <select>
        const select = container.element(by.css("param-link mat-select"));
        return select;
    }

    /**
     * Returns an object containing all the calculator's inputs values, indexed
     * by parameter ID
     */
    async storeAllInputValues() {
        const inputs = this.getParamInputs();
        const values = {};
        await inputs.each(async (i) => {
            const inputId = await i.getAttribute("id");
            const inputValue = await i.getAttribute("value");
            values[inputId] = +inputValue; // cast to number to avoid false negative (integers starting with 0)
        });
        return values;
    }

    /**
     * Modifies all the calculator's editable inputs values by adding a random digit other than 0 at the end
     */
    async modifyAllInputValues() {
        const inputs = this.getParamInputs();
        await inputs.each(async (i) => {
            if (await i.isDisplayed()) {
                // N in YAXN child of SPP module must not be float
                const isN = (await i.getAttribute("id")).includes("_N"); // @TODO strengthen this clodo test
                // Ob in Grille is set to 0.5 but cannot exceed 0.58; do not touch it
                const isOb = (await i.getAttribute("id")) === "Ob";
                const hasDot = (await i.getAttribute("value")).includes(".");
                const hasExponent = (await i.getAttribute("value")).includes("e");
                let keys = "" + Math.floor(Math.random() * 9) + 1;
                if (!hasDot && !hasExponent && !isN) {
                    keys = "." + keys;
                }
                if (! isOb && await i.getAttribute("disabled") === null) {
                    await i.sendKeys(keys);
                }
            }
        });
    }
}
